"""The main Python package."""

from .build_scripts import build
from xrpl_plugin import basic_types
from xrpl_plugin import models
from xrpl_plugin import nfts
from xrpl_plugin import sfields
from xrpl_plugin import stypes
from xrpl_plugin import return_codes
from xrpl_plugin import keylets
from xrpl_plugin import ledger_objects
from xrpl_plugin import transactors

__all__ = [
    "models",
    "sfields",
    "stypes",
    "return_codes",
    "basic_types",
    "keylets",
    "ledger_objects",
    "transactors",
    "build",
    "nfts",
]



from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout
import re


class XrplPlugin(ConanFile):
    name = "xrpl-plugin"

    license = "ISC"
    url = "https://github.com/mvadari/xrpl-plugin"
    description = "Plugins for the XRP Ledger"
    settings = "os", "compiler", "build_type", "arch"
    options = {
        "assertions": [True, False],
        "coverage": [True, False],
        "fPIC": [True, False],
        "jemalloc": [True, False],
        "reporting": [True, False],
        "rocksdb": [True, False],
        "shared": [True, False],
        "static": [True, False],
        "tests": [True, False],
        "unity": [True, False],
    }

    requires = [
        "boost/1.82.0",
        "date/3.0.1",
        "grpc/1.50.1",
        "libarchive/3.6.2",
        "lz4/1.9.3",
        "nudb/2.0.8",
        "openssl/1.1.1u",
        "protobuf/3.21.9",
        "pybind11/2.10.4",
        "snappy/1.1.10",
        "soci/4.0.3",
        "sqlite3/3.42.0",
        "zlib/1.2.13",
    ]

    default_options = {
        "assertions": False,
        "coverage": False,
        "fPIC": True,
        "jemalloc": False,
        "reporting": False,
        "rocksdb": True,
        "shared": False,
        "static": True,
        "tests": True,
        "unity": False,
        "cassandra-cpp-driver:shared": False,
        "cassandra-cpp-driver:use_atomic": None,
        "date:header_only": True,
        "grpc:shared": False,
        "grpc:secure": True,
        "libarchive:shared": False,
        "libarchive:with_acl": False,
        "libarchive:with_bzip2": False,
        "libarchive:with_cng": False,
        "libarchive:with_expat": False,
        "libarchive:with_iconv": False,
        "libarchive:with_libxml2": False,
        "libarchive:with_lz4": True,
        "libarchive:with_lzma": False,
        "libarchive:with_lzo": False,
        "libarchive:with_nettle": False,
        "libarchive:with_openssl": False,
        "libarchive:with_pcreposix": False,
        "libarchive:with_xattr": False,
        "libarchive:with_zlib": False,
        "libpq:shared": False,
        "lz4:shared": False,
        "openssl:shared": False,
        "protobuf:shared": False,
        "protobuf:with_zlib": True,
        "rocksdb:enable_sse": False,
        "rocksdb:lite": False,
        "rocksdb:shared": False,
        "rocksdb:use_rtti": True,
        "rocksdb:with_jemalloc": False,
        "rocksdb:with_lz4": True,
        "rocksdb:with_snappy": True,
        "snappy:shared": False,
        "soci:shared": False,
        "soci:with_sqlite3": True,
        "soci:with_boost": True,
    }

    def configure(self):
        if self.settings.compiler == "apple-clang":
            self.options["boost"].visibility = "global"

    def requirements(self):
        if self.options.jemalloc:
            self.requires("jemalloc/5.3.0")
        if self.options.reporting:
            self.requires("cassandra-cpp-driver/2.15.3")
            self.requires("libpq/14.7")
        if self.options.rocksdb:
            self.requires("rocksdb/6.29.5")

    exports_sources = (
        "CMakeLists.txt",
        "Builds/*",
        "bin/getRippledInfo",
        "src/*",
        "cfg/*",
    )

    def layout(self):
        cmake_layout(self)
        # Fix this setting to follow the default introduced in Conan 1.48
        # to align with our build instructions.
        self.folders.generators = "build/generators"

    generators = "CMakeDeps"

    def generate(self):
        tc = CMakeToolchain(self)
        tc.variables["tests"] = self.options.tests
        tc.variables["assert"] = self.options.assertions
        tc.variables["coverage"] = self.options.coverage
        tc.variables["jemalloc"] = self.options.jemalloc
        tc.variables["reporting"] = self.options.reporting
        tc.variables["rocksdb"] = self.options.rocksdb
        tc.variables["BUILD_SHARED_LIBS"] = self.options.shared
        tc.variables["static"] = self.options.static
        tc.variables["unity"] = self.options.unity
        tc.generate()

    def build(self):
        cmake = CMake(self)
        cmake.verbose = True
        cmake.configure()
        cmake.build()

    def package(self):
        cmake = CMake(self)
        cmake.verbose = True
        cmake.install()

    def package_info(self):
        libxrpl = self.cpp_info.components["libxrpl"]
        libxrpl.libs = [
            "libxrpl_core.a",
            "libed25519.a",
            "libsecp256k1.a",
        ]
        # TODO: Fix the protobufs to include each other relative to
        # `include/`, not `include/ripple/proto/`.
        libxrpl.includedirs = ["include", "include/ripple/proto"]
        libxrpl.requires = [
            "boost::boost",
            "openssl::crypto",
            "date::date",
            "grpc::grpc++",
        ]



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple, Union

from xrpl_plugin.transactors import (
    Application,
    ApplyContext,
    ApplyView,
    ConsequencesFactoryType,
    PreclaimContext,
    PreflightContext,
    ReadView,
    SerialIter,
    Serializer,
    TxConsequences,
)
from xrpl_plugin.basic_types import (
    AccountID,
    Buffer,
    Journal,
    JsonValue,
    VoteBehavior,
    XRPAmount,
    uint256,
    SOEStyle,
)
from xrpl_plugin.sfields import SField
from xrpl_plugin.stypes import (
    STLedgerEntry,
    STTx,
)
from xrpl_plugin.return_codes import TER, NotTEC


@dataclass(frozen=True)
class Transactor:
    name: str
    tx_type: int
    tx_format: List[Tuple[SField, SOEStyle]]
    consequences_factory_type: ConsequencesFactoryType
    make_tx_consequences: Optional[Callable[[PreflightContext], TxConsequences]] = None
    calculate_base_fee: Optional[Callable[[ReadView, STTx], XRPAmount]] = None
    preflight: Optional[Callable[[PreflightContext], NotTEC]] = None
    preclaim: Optional[Callable[[PreclaimContext], TER]] = None
    do_apply: Optional[Callable[[ApplyContext, XRPAmount, XRPAmount], TER]] = None
    check_seq_proxy: Optional[Callable[[ReadView, STTx, Journal], NotTEC]] = None
    check_prior_tx_and_last_ledger: Optional[Callable[[PreclaimContext], NotTEC]] = None
    check_fee: Optional[Callable[[PreclaimContext, XRPAmount], TER]] = None
    check_sign: Optional[Callable[[PreclaimContext], TER]] = None


@dataclass(frozen=True)
class SType:
    type_id: int
    parse_value: Callable[
        [SField, str, str, SField, JsonValue], Union[Buffer, JsonValue]
    ]
    to_string: Callable[[Buffer], str]
    to_serializer: Callable[[Buffer, Serializer], None]
    from_serial_iter: Callable[[SerialIter], Buffer]
    to_json: Optional[Callable[[Buffer], JsonValue]] = None


@dataclass(frozen=True)
class Amendment:
    name: str
    supported: bool
    vote: VoteBehavior


@dataclass(frozen=True)
class TERCode:
    code: int
    code_str: str
    description: str


class InvariantCheck(ABC):
    def process_data(self, data):
        if data is None:
            return
        data_dict = json.loads(data)
        self.__dict__.update(**data_dict)

    def visit_entry_actual(self, data, is_delete, before, after):
        self.process_data(data)

        self.visit_entry(is_delete, before, after)

        return json.dumps(self.__dict__)

    def finalize_actual(self, data, tx, result, fee, view, j):
        self.process_data(data)
        return self.finalize(tx, result, fee, view, j)

    @abstractmethod
    def visit_entry(self, is_delete, before, after):
        pass

    @abstractmethod
    def finalize(self, tx, result, fee, view, j):
        pass


@dataclass(frozen=True)
class LedgerObject:
    object_type: int
    name: str
    rpc_name: str
    object_format: List[Tuple[SField, SOEStyle]]
    is_deletion_blocker: bool = False
    delete_object: Optional[
        Callable[
            [Application, ApplyView, AccountID, uint256, STLedgerEntry, Journal], TER
        ]
    ] = None
    visit_entry_xrp_change: Optional[Callable[[bool, STLedgerEntry, bool], int]] = None

    def __post_init__(self):
        if self.is_deletion_blocker and self.delete_object is not None:
            raise Exception(
                "Cannot have `delete_object` if `is_deletion_blocker` is True"
            )
        if not self.is_deletion_blocker and self.delete_object is None:
            raise Exception(
                "Must have `delete_object` if `is_deletion_blocker` is False"
            )


@dataclass(frozen=True)
class InnerObject:
    field: SField
    format: List[Tuple[SField, SOEStyle]]



from conan import ConanFile
from conan.tools.build import check_min_cppstd
from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout
from conan.tools.files import apply_conandata_patches, copy, export_conandata_patches, get, rmdir
from conan.tools.scm import Version
import os

required_conan_version = ">=1.54.0"


class SnappyConan(ConanFile):
    name = "snappy"
    description = "A fast compressor/decompressor"
    topics = ("google", "compressor", "decompressor")
    url = "https://github.com/conan-io/conan-center-index"
    homepage = "https://github.com/google/snappy"
    license = "BSD-3-Clause"

    package_type = "library"
    settings = "os", "arch", "compiler", "build_type"
    options = {
        "shared": [True, False],
        "fPIC": [True, False],
    }
    default_options = {
        "shared": False,
        "fPIC": True,
    }

    def export_sources(self):
        export_conandata_patches(self)

    def config_options(self):
        if self.settings.os == 'Windows':
            del self.options.fPIC

    def configure(self):
        if self.options.shared:
            self.options.rm_safe("fPIC")

    def layout(self):
        cmake_layout(self, src_folder="src")

    def validate(self):
        if self.settings.compiler.get_safe("cppstd"):
            check_min_cppstd(self, 11)

    def source(self):
        get(self, **self.conan_data["sources"][self.version], strip_root=True)

    def generate(self):
        tc = CMakeToolchain(self)
        tc.variables["SNAPPY_BUILD_TESTS"] = False
        if Version(self.version) >= "1.1.8":
            tc.variables["SNAPPY_FUZZING_BUILD"] = False
            tc.variables["SNAPPY_REQUIRE_AVX"] = False
            tc.variables["SNAPPY_REQUIRE_AVX2"] = False
            tc.variables["SNAPPY_INSTALL"] = True
        if Version(self.version) >= "1.1.9":
            tc.variables["SNAPPY_BUILD_BENCHMARKS"] = False
        tc.generate()

    def build(self):
        apply_conandata_patches(self)
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def package(self):
        copy(self, "COPYING", src=self.source_folder, dst=os.path.join(self.package_folder, "licenses"))
        cmake = CMake(self)
        cmake.install()
        rmdir(self, os.path.join(self.package_folder, "lib", "cmake"))

    def package_info(self):
        self.cpp_info.set_property("cmake_file_name", "Snappy")
        self.cpp_info.set_property("cmake_target_name", "Snappy::snappy")
        # TODO: back to global scope in conan v2 once cmake_find_package* generators removed
        self.cpp_info.components["snappylib"].libs = ["snappy"]
        if not self.options.shared:
            if self.settings.os in ["Linux", "FreeBSD"]:
                self.cpp_info.components["snappylib"].system_libs.append("m")

        # TODO: to remove in conan v2 once cmake_find_package* generators removed
        self.cpp_info.names["cmake_find_package"] = "Snappy"
        self.cpp_info.names["cmake_find_package_multi"] = "Snappy"
        self.cpp_info.components["snappylib"].names["cmake_find_package"] = "snappy"
        self.cpp_info.components["snappylib"].names["cmake_find_package_multi"] = "snappy"
        self.cpp_info.components["snappylib"].set_property("cmake_target_name", "Snappy::snappy")



import xrpl_plugin.rippled_py.return_codes as rippled_return_codes
from xrpl_plugin.rippled_py.return_codes import *

__doc__ = rippled_return_codes.__doc__

__all__ = list(rippled_return_codes.__dict__.keys())



import xrpl_plugin.rippled_py.transactors as rippled_transactors
from xrpl_plugin.rippled_py.transactors import *

__doc__ = rippled_transactors.__doc__

__all__ = list(rippled_transactors.__dict__.keys())



import xrpl_plugin.rippled_py.basic_types as rippled_basic_types
from xrpl_plugin.rippled_py.basic_types import *

__doc__ = rippled_basic_types.__doc__

__all__ = list(rippled_basic_types.__dict__.keys())



from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, cmake_layout
import re

class Xrpl(ConanFile):
    name = 'xrpl'

    license = 'ISC'
    author = 'John Freeman <jfreeman@ripple.com>'
    url = 'https://github.com/xrplf/rippled'
    description = 'The XRP Ledger'
    settings = 'os', 'compiler', 'build_type', 'arch'
    options = {
        'assertions': [True, False],
        'coverage': [True, False],
        'fPIC': [True, False],
        'jemalloc': [True, False],
        'reporting': [True, False],
        'rocksdb': [True, False],
        'shared': [True, False],
        'static': [True, False],
        'tests': [True, False],
        'unity': [True, False],
    }

    requires = [
        'boost/1.82.0',
        'date/3.0.1',
        'grpc/1.50.1',
        'libarchive/3.6.2',
        'lz4/1.9.3',
        'nudb/2.0.8',
        'openssl/1.1.1u',
        'protobuf/3.21.9',
        'snappy/1.1.10',
        'soci/4.0.3',
        'sqlite3/3.42.0',
        'zlib/1.2.13',
    ]

    default_options = {
        'assertions': False,
        'coverage': False,
        'fPIC': True,
        'jemalloc': False,
        'reporting': False,
        'rocksdb': True,
        'shared': False,
        'static': True,
        'tests': True,
        'unity': False,

        'cassandra-cpp-driver:shared': False,
        'cassandra-cpp-driver:use_atomic': None,
        'date:header_only': True,
        'grpc:shared': False,
        'grpc:secure': True,
        'libarchive:shared': False,
        'libarchive:with_acl': False,
        'libarchive:with_bzip2': False,
        'libarchive:with_cng': False,
        'libarchive:with_expat': False,
        'libarchive:with_iconv': False,
        'libarchive:with_libxml2': False,
        'libarchive:with_lz4': True,
        'libarchive:with_lzma': False,
        'libarchive:with_lzo': False,
        'libarchive:with_nettle': False,
        'libarchive:with_openssl': False,
        'libarchive:with_pcreposix': False,
        'libarchive:with_xattr': False,
        'libarchive:with_zlib': False,
        'libpq:shared': False,
        'lz4:shared': False,
        'openssl:shared': False,
        'protobuf:shared': False,
        'protobuf:with_zlib': True,
        'rocksdb:enable_sse': False,
        'rocksdb:lite': False,
        'rocksdb:shared': False,
        'rocksdb:use_rtti': True,
        'rocksdb:with_jemalloc': False,
        'rocksdb:with_lz4': True,
        'rocksdb:with_snappy': True,
        'snappy:shared': False,
        'soci:shared': False,
        'soci:with_sqlite3': True,
        'soci:with_boost': True,
    }

    def set_version(self):
        path = f'{self.recipe_folder}/src/ripple/protocol/impl/BuildInfo.cpp'
        regex = r'versionString\s?=\s?\"(.*)\"'
        with open(path, 'r') as file:
            matches = (re.search(regex, line) for line in file)
            match = next(m for m in matches if m)
            self.version = match.group(1)

    def configure(self):
        if self.settings.compiler == 'apple-clang':
            self.options['boost'].visibility = 'global'

    def requirements(self):
        if self.options.jemalloc:
            self.requires('jemalloc/5.3.0')
        if self.options.reporting:
            self.requires('cassandra-cpp-driver/2.15.3')
            self.requires('libpq/14.7')
        if self.options.rocksdb:
            self.requires('rocksdb/6.29.5')

    exports_sources = (
        'CMakeLists.txt', 'Builds/*', 'bin/getRippledInfo', 'src/*', 'cfg/*'
    )

    def layout(self):
        cmake_layout(self)
        # Fix this setting to follow the default introduced in Conan 1.48
        # to align with our build instructions.
        self.folders.generators = 'build/generators'

    generators = 'CMakeDeps'
    def generate(self):
        tc = CMakeToolchain(self)
        tc.variables['tests'] = self.options.tests
        tc.variables['assert'] = self.options.assertions
        tc.variables['coverage'] = self.options.coverage
        tc.variables['jemalloc'] = self.options.jemalloc
        tc.variables['reporting'] = self.options.reporting
        tc.variables['rocksdb'] = self.options.rocksdb
        tc.variables['BUILD_SHARED_LIBS'] = self.options.shared
        tc.variables['static'] = self.options.static
        tc.variables['unity'] = self.options.unity
        tc.generate()

    def build(self):
        cmake = CMake(self)
        cmake.verbose = True
        cmake.configure()
        cmake.build()

    def package(self):
        cmake = CMake(self)
        cmake.verbose = True
        cmake.install()

    def package_info(self):
        libxrpl = self.cpp_info.components['libxrpl']
        libxrpl.libs = [
            'libxrpl_core.a',
            'libed25519.a',
            'libsecp256k1.a',
        ]
        # TODO: Fix the protobufs to include each other relative to
        # `include/`, not `include/ripple/proto/`.
        libxrpl.includedirs = ['include', 'include/ripple/proto']
        libxrpl.requires = [
            'boost::boost',
            'openssl::crypto',
            'date::date',
            'grpc::grpc++',
        ]



import os
import sys
import subprocess
import tempfile
from pathlib import Path


def generate_cpp(tx_name, module_name, python_folder):
    return f"""#include <ripple/app/tx/TxConsequences.h>
#include <ripple/basics/Log.h>
#include <ripple/basics/XRPAmount.h>
#include <ripple/ledger/ApplyView.h>
#include <ripple/ledger/View.h>
#include <ripple/plugin/exports.h>
#include <ripple/protocol/Feature.h>
#include <ripple/protocol/TER.h>
#include <ripple/protocol/TxFlags.h>
#include <ripple/protocol/st.h>
#include <map>
#include <vector>

#include <pybind11/embed.h> // everything needed for embedding
#include <pybind11/stl.h>

namespace py = pybind11;
using namespace pybind11::literals; // to bring in the `_a` literal

using namespace ripple;

struct WSF {{
  void *f_;
  operator SField const &() const {{
    return *static_cast<SField *>(f_);
  }};
}};

static int interpreterCount = 0;

class CustomScopedInterpreter
{{
public:
    explicit CustomScopedInterpreter()
    {{
        if (interpreterCount == 0) py::initialize_interpreter();
        interpreterCount += 1;
    }}

    ~CustomScopedInterpreter()
    {{
        interpreterCount -= 1;
        if (interpreterCount == 0) py::finalize_interpreter();
    }}
}};

// ----------------------------------------------------------------------------
// Transactors
// ----------------------------------------------------------------------------

static std::map<std::uint16_t, int> txTypeToIndexMap;

py::object
getTransactorFromPython(std::uint16_t txType)
{{
    py::module_::import("sys").attr("path").attr("append")("{python_folder}");
    py::module module = py::module_::import("{module_name}");
    if (auto it = txTypeToIndexMap.find(txType);
        it != txTypeToIndexMap.end())
    {{
        int index = it->second;
        return module.attr("transactors")[py::cast(index)];
    }}
    std::string errorMessage = "Cannot find tx type value " + std::to_string(txType);
    std::cout << "Python Processing Error: " << errorMessage << std::endl;
    throw std::runtime_error(errorMessage);
}}

py::object
getTransactorFromPython(std::uint16_t txType, beast::Journal const j)
{{
    py::module_::import("sys").attr("path").attr("append")("{python_folder}");
    py::module module = py::module_::import("{module_name}");
    if (auto it = txTypeToIndexMap.find(txType);
        it != txTypeToIndexMap.end())
    {{
        int index = it->second;
        return module.attr("transactors")[py::cast(index)];
    }}
    std::string errorMessage = "Cannot find tx type value " + std::to_string(txType);
    JLOG(j.error()) << "Python Processing Error: " << errorMessage;
    throw std::runtime_error(errorMessage);
}}

TxConsequences
makeTxConsequences(PreflightContext const& ctx)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.j);
    py::object function = transactor.attr("makeTxConsequences");
    try {{
        py::object ret = function(py::cast(ctx, py::return_value_policy::reference));
        return ret.cast<TxConsequences>();
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

XRPAmount
calculateBaseFee(ReadView const& view, STTx const& tx)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(tx.getTxnType());
    py::object function = transactor.attr("calculateBaseFee");
    try {{
        py::object ret = function(
            py::cast(view, py::return_value_policy::reference),
            py::cast(tx, py::return_value_policy::reference)
        );
        return ret.cast<XRPAmount>();
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

NotTEC
preflight(PreflightContext const& ctx)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.j);
    py::object function = transactor.attr("preflight");
    try {{
        py::object ret = function(py::cast(ctx, py::return_value_policy::reference));
        try {{
            return ret.cast<NotTEC>();
        }} catch (const py::cast_error &) {{ // TODO: figure out the exact error that is thrown
            return NotTEC::fromInt(ret.cast<int>());
        }}
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

TER
preclaim(PreclaimContext const& ctx)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.j);
    py::object function = transactor.attr("preclaim");
    try {{
        py::object ret = function(py::cast(ctx, py::return_value_policy::reference));
        return TER::fromInt(ret.cast<int>());
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

TER
doApply(ApplyContext& ctx, XRPAmount mPriorBalance, XRPAmount mSourceBalance)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.journal);
    py::object function = transactor.attr("do_apply");
    try {{
        py::object ret = function(
            py::cast(ctx, py::return_value_policy::reference),
            mPriorBalance,
            mSourceBalance);
        return TER::fromInt(ret.cast<int>());
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.journal.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

NotTEC
checkSeqProxy(ReadView const& view, STTx const& tx, beast::Journal j)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(tx.getTxnType(), j);
    py::object function = transactor.attr("check_seq_proxy");
    try {{
        py::object ret = function(
            py::cast(view, py::return_value_policy::reference),
            py::cast(tx, py::return_value_policy::reference),
            py::cast(j, py::return_value_policy::reference)
        );
        try {{
            return ret.cast<NotTEC>();
        }} catch (const py::cast_error &) {{ // TODO: figure out the exact error that is thrown
            return NotTEC::fromInt(ret.cast<int>());
        }}
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

NotTEC
checkPriorTxAndLastLedger(PreclaimContext const& ctx)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.j);
    py::object function = transactor.attr("check_prior_tx_and_last_ledger_ptr");
    try {{
        py::object functionReturn = function(py::cast(ctx, py::return_value_policy::reference));
        return NotTEC::fromInt(functionReturn.cast<int>());
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

TER
checkFee(PreclaimContext const& ctx, XRPAmount baseFee)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.j);
    py::object function = transactor.attr("check_fee");
    try {{
        py::object ret = function(
            py::cast(ctx, py::return_value_policy::reference),
            baseFee
        );
        return TER::fromInt(ret.cast<int>());
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

NotTEC
checkSign(PreclaimContext const& ctx)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object transactor = getTransactorFromPython(ctx.tx.getTxnType(), ctx.j);
    py::object function = transactor.attr("check_sign");
    try {{
        py::object ret = function(py::cast(ctx, py::return_value_policy::reference));
        return NotTEC::fromInt(ret.cast<int>());
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        JLOG(ctx.j.error()) << "Python Error: " << errorMessage;
        throw;
    }}
}}

struct TransactorExportInternal {{
    std::string txName;
    std::uint16_t txType;
    std::vector<SOElementExport> txFormat;
    ConsequencesFactoryType consequencesFactoryType;
    makeTxConsequencesPtr makeTxConsequences = nullptr;
    calculateBaseFeePtr calculateBaseFee = nullptr;
    preflightPtr preflight = nullptr;
    preclaimPtr preclaim = nullptr;
    doApplyPtr doApply = nullptr;
    checkSeqProxyPtr checkSeqProxy = nullptr;
    checkPriorTxAndLastLedgerPtr checkPriorTxAndLastLedger = nullptr;
    checkFeePtr checkFee = nullptr;
    checkSignPtr checkSign = nullptr;
}};

TransactorExport
mutateTransactor(TransactorExportInternal const& tx)
{{
    return TransactorExport{{
        tx.txName.c_str(),
        tx.txType,
        {{const_cast<SOElementExport *>(tx.txFormat.data()), static_cast<int>(tx.txFormat.size())}},
        tx.consequencesFactoryType,
        tx.makeTxConsequences,
        tx.calculateBaseFee,
        tx.preflight,
        tx.preclaim,
        tx.doApply,
        tx.checkSeqProxy,
        tx.checkPriorTxAndLastLedger,
        tx.checkFee,
        tx.checkSign,
    }};
}}

extern "C"
Container<TransactorExport>
getTransactors()
{{
    static std::vector<TransactorExportInternal> const transactors = []{{
        std::vector<TransactorExportInternal> temp = {{}};
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module pluginImport = py::module_::import("{module_name}");
        if (!hasattr(pluginImport, "transactors")) {{
            return temp;
        }}
        auto transactors = pluginImport.attr("transactors").cast<std::vector<py::object>>();
        for (int i = 0; i < transactors.size(); i++)
        {{
            py::object transactor = transactors[i];
            auto txType = transactor.attr("tx_type").cast<std::uint16_t>();
            txTypeToIndexMap.insert({{txType, i}});

            std::vector<SOElementExport> format{{}};
            auto pythonFormat = transactor.attr("tx_format").cast<std::vector<py::object>>();
            for (py::object transactor: pythonFormat)
            {{
                py::tuple tup = transactor.cast<py::tuple>();
                WSF sfield = tup[0].cast<WSF>();
                SOEStyle varType = tup[1].cast<SOEStyle>();
                format.emplace_back(SOElementExport{{
                    static_cast<SField const&>(sfield).getCode(),
                    varType}});
            }}

            auto exportedTx = TransactorExportInternal{{
                transactor.attr("name").cast<std::string>(),
                txType,
                format,
                transactor.attr("consequences_factory_type").cast<ConsequencesFactoryType>(),
                transactor.attr("make_tx_consequences").is_none() ? nullptr : makeTxConsequences,
                transactor.attr("calculate_base_fee").is_none() ? nullptr : calculateBaseFee,
                transactor.attr("preflight").is_none() ? nullptr : preflight,
                transactor.attr("preclaim").is_none() ? nullptr : preclaim,
                transactor.attr("do_apply").is_none() ? nullptr : doApply,
                transactor.attr("check_seq_proxy").is_none() ? nullptr : checkSeqProxy,
                transactor.attr("check_prior_tx_and_last_ledger").is_none() ? nullptr : checkPriorTxAndLastLedger,
                transactor.attr("check_fee").is_none() ? nullptr : checkFee,
                transactor.attr("check_sign").is_none() ? nullptr : checkSign,
            }};

            temp.emplace_back(exportedTx);
        }}
        return temp;
    }}();
    static std::vector<TransactorExport> output;
    output.reserve(transactors.size());
    std::transform(transactors.begin(), transactors.end(), std::back_inserter(output), mutateTransactor);
    return {{const_cast<TransactorExport *>(output.data()), static_cast<int>(output.size())}};
}}

// ----------------------------------------------------------------------------
// Ledger Objects
// ----------------------------------------------------------------------------

static std::map<std::uint16_t, int> ledgerObjectTypeToIndexMap;

struct LedgerObjectExportInternal {{
    std::uint16_t type;
    std::string name; // CamelCase
    std::string rpcName; // snake_case
    std::vector<SOElementExport> format;
    bool isDeletionBlocker;
    DeleterFuncPtr deleteFn;
    visitEntryXRPChangePtr visitEntryXRPChange;
}};

LedgerObjectExport
mutate(LedgerObjectExportInternal const& obj)
{{
    return LedgerObjectExport{{
        obj.type,
        obj.name.c_str(),
        obj.rpcName.c_str(),
        {{const_cast<SOElementExport *>(obj.format.data()), static_cast<int>(obj.format.size())}},
        obj.isDeletionBlocker,
        obj.deleteFn,
        obj.visitEntryXRPChange,
    }};
}}

py::object
getLedgerObjectFromPython(std::uint16_t object)
{{
    py::module_::import("sys").attr("path").attr("append")("{python_folder}");
    py::module module = py::module_::import("{module_name}");
    if (auto it = ledgerObjectTypeToIndexMap.find(object);
        it != ledgerObjectTypeToIndexMap.end())
    {{
        int index = it->second;
        return module.attr("ledger_objects")[py::cast(index)];
    }}
    std::string errorMessage = "Cannot find ledger object type value " + std::to_string(object);
    std::cout << "Python Processing Error: " << errorMessage << std::endl;
    throw std::runtime_error(errorMessage);
}}

TER
deleteObject(
    Application& app,
    ApplyView& view,
    AccountID const& account,
    uint256 const& delIndex,
    std::shared_ptr<SLE> const& sleDel,
    beast::Journal j)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getLedgerObjectFromPython(sleDel->getType());
    py::object function = object.attr("delete_object");
    try {{
        py::object ret = function(
            py::cast(app, py::return_value_policy::reference),
            py::cast(view, py::return_value_policy::reference),
            py::cast(account, py::return_value_policy::reference),
            delIndex,
            py::cast(sleDel, py::return_value_policy::reference),
            py::cast(j, py::return_value_policy::reference)
        );
        return TER::fromInt(ret.cast<int>());
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

std::int64_t
visitEntryXRPChange(
    bool isDelete,
    std::shared_ptr<STLedgerEntry const> const& entry,
    bool isBefore)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getLedgerObjectFromPython(entry->getType());
    py::object function = object.attr("visit_entry_xrp_change");
    try {{
        py::object ret = function(
            isDelete,
            py::cast(entry, py::return_value_policy::reference),
            isBefore
        );
        return ret.cast<std::int64_t>();
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

extern "C"
Container<LedgerObjectExport>
getLedgerObjects()
{{
    static std::vector<LedgerObjectExportInternal> const objects = []{{
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module_ module = py::module_::import("{module_name}");
        std::vector<LedgerObjectExportInternal> temp = {{}};
        if (!hasattr(module, "ledger_objects")) {{
            return temp;
        }}
        auto objects = module.attr("ledger_objects").cast<std::vector<py::object>>();
        for (int i = 0; i < objects.size(); i++)
        {{
            py::object object = objects[i];
            auto objectType = object.attr("object_type").cast<std::uint16_t>();
            ledgerObjectTypeToIndexMap.insert({{objectType, i}});

            std::vector<SOElementExport> objectFormat{{}};
            auto pythonObjectFormat = object.attr("object_format").cast<std::vector<py::object>>();
            for (py::object variable: pythonObjectFormat)
            {{
                py::tuple tup = variable.cast<py::tuple>();
                WSF sfield = tup[0].cast<WSF>();
                SOEStyle varType = tup[1].cast<SOEStyle>();
                objectFormat.emplace_back(SOElementExport{{
                    static_cast<SField const&>(sfield).getCode(),
                    varType}});
            }}
            temp.emplace_back(LedgerObjectExportInternal{{
                objectType,
                object.attr("name").cast<std::string>(),
                object.attr("rpc_name").cast<std::string>(),
                std::move(objectFormat),
                object.attr("is_deletion_blocker").cast<bool>(),
                object.attr("delete_object").is_none() ? nullptr : deleteObject,
                object.attr("visit_entry_xrp_change").is_none() ? nullptr : visitEntryXRPChange}});
        }}
        return temp;
    }}();
    static std::vector<LedgerObjectExport> output;
    output.reserve(objects.size());
    std::transform(objects.begin(), objects.end(), std::back_inserter(output), mutate);
    return {{const_cast<LedgerObjectExport *>(output.data()), static_cast<int>(output.size())}};
}}

// ----------------------------------------------------------------------------
// SFields
// ----------------------------------------------------------------------------

static struct
{{
    bool operator()(py::object a, py::object b) const {{ 
        return a.attr("fieldNum").cast<int>() < b.attr("fieldNum").cast<int>();
    }}
}}
sFieldSorter;

extern "C"
Container<SFieldExport>
getSFields()
{{
    static std::vector<SFieldExport> const sFields = []{{
        std::vector<SFieldExport> temp = {{}};
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module pluginImport = py::module_::import("{module_name}");
        if (!hasattr(pluginImport, "sfields")) {{
            return temp;
        }}
        auto sfields = pluginImport.attr("sfields").cast<std::vector<py::object>>();
        std::sort(sfields.begin(), sfields.end(), sFieldSorter);
        for (py::object& variable: sfields)
        {{
            WSF wrappedSField = variable.cast<WSF>();
            SField const& sfield = static_cast<SField const&>(wrappedSField);
            temp.emplace_back(SFieldExport{{
                sfield.fieldType,
                sfield.fieldValue,
                sfield.jsonName}});
        }}
        return temp;
    }}();
    return {{const_cast<SFieldExport *>(sFields.data()), static_cast<int>(sFields.size())}};
}}

// ----------------------------------------------------------------------------
// Serialized Types
// ----------------------------------------------------------------------------

static std::map<int, int> sTypeToIndexMap;

py::object
getSTypeFromPython(std::uint16_t object)
{{
    py::module_::import("sys").attr("path").attr("append")("{python_folder}");
    py::module module = py::module_::import("{module_name}");
    if (auto it = sTypeToIndexMap.find(object);
        it != sTypeToIndexMap.end())
    {{
        int index = it->second;
        return module.attr("stypes")[py::cast(index)];
    }}
    std::string errorMessage = "Cannot find stype value " + std::to_string(object);
    std::cout << "Python Processing Error: " << errorMessage << std::endl;
    throw std::runtime_error(errorMessage);
}}

Buffer
parseValue(
    SField const& field,
    std::string const& json_name,
    std::string const& fieldName,
    SField const* name,
    Json::Value const& value,
    Json::Value& error)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getSTypeFromPython(field.fieldType);
    py::object function = object.attr("parse_value");
    try {{
        WSF wrappedField = WSF{{(void *)&field}};
        WSF wrappedName = WSF{{(void *)name}};
        py::object returnObj = function(
            py::cast(wrappedField, py::return_value_policy::reference),
            json_name,
            fieldName,
            wrappedName,
            value);
        try {{
            return returnObj.cast<Buffer>();
        }} catch (const py::cast_error &) {{ // TODO: figure out the exact error that is thrown
            error = returnObj.cast<Json::Value>();
            return Buffer();
        }}
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

std::string
toString(int typeId, Buffer const& buf)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getSTypeFromPython(typeId);
    py::object function = object.attr("to_string");
    try {{
        py::object ret = function(py::cast(buf, py::return_value_policy::reference));
        return ret.cast<std::string>();
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

Json::Value
toJson(int typeId, Buffer const& buf)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getSTypeFromPython(typeId);
    py::object function = object.attr("to_json");
    try {{
        py::object ret = function(py::cast(buf, py::return_value_policy::reference));
        return ret.cast<Json::Value>();
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

void
toSerializer(int typeId, Buffer const& buf, Serializer& s)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getSTypeFromPython(typeId);
    py::object function = object.attr("to_serializer");
    try {{
        function(
            py::cast(buf, py::return_value_policy::reference),
            py::cast(s, py::return_value_policy::reference)
        );
        return;
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

Buffer
fromSerialIter(int typeId, SerialIter& st)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::object object = getSTypeFromPython(typeId);
    py::object function = object.attr("from_serial_iter");
    try {{
        py::object ret = function(py::cast(st, py::return_value_policy::reference));
        return ret.cast<Buffer>();
    }} catch (py::error_already_set& e) {{
        // Print the error message
        const char* errorMessage = e.what();
        std::cout << "Python Error: " << errorMessage << std::endl;
        throw;
    }}
}}

extern "C"
Container<STypeExport>
getSTypes()
{{
    static std::vector<STypeExport> const types = []{{
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module_ module = py::module_::import("{module_name}");
        std::vector<STypeExport> temp = {{}};
        if (!hasattr(module, "stypes")) {{
            return temp;
        }}
        auto types = module.attr("stypes").cast<std::vector<py::object>>();
        for (int i = 0; i < types.size(); i++)
        {{
            py::object stype = types[i];
            auto typeId = stype.attr("type_id").cast<int>();
            sTypeToIndexMap.insert({{typeId, i}});

            temp.emplace_back(STypeExport{{
                typeId,
                parseValue,
                toString,
                stype.attr("to_json").is_none() ? nullptr : toJson,
                toSerializer,
                fromSerialIter}});
        }}
        return temp;
    }}();
    return {{const_cast<STypeExport *>(types.data()), static_cast<int>(types.size())}};
}}

// ----------------------------------------------------------------------------
// Amendments
// ----------------------------------------------------------------------------

struct AmendmentExportInternal {{
    std::string name;
    bool supported;
    VoteBehavior vote;
}};

AmendmentExport
mutateAmendment(AmendmentExportInternal const& amendment)
{{
    return AmendmentExport{{
        amendment.name.c_str(),
        amendment.supported,
        amendment.vote,
    }};
}}

extern "C"
Container<AmendmentExport>
getAmendments()
{{
    static std::vector<AmendmentExportInternal> const amendments = []{{
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        std::vector<AmendmentExportInternal> temp = {{}};
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module pluginImport = py::module_::import("{module_name}");
        if (!hasattr(pluginImport, "amendments")) {{
            return temp;
        }}
        auto amendments = pluginImport.attr("amendments").cast<std::vector<py::object>>();
        for (py::object variable: amendments)
        {{
            temp.emplace_back(AmendmentExportInternal{{
                variable.attr("name").cast<std::string>(),
                variable.attr("supported").cast<bool>(),
                variable.attr("vote").cast<VoteBehavior>()}});
        }}
        return temp;
    }}();
    static std::vector<AmendmentExport> output;
    output.reserve(amendments.size());
    std::transform(amendments.begin(), amendments.end(), std::back_inserter(output), mutateAmendment);
    return {{const_cast<AmendmentExport *>(output.data()), static_cast<int>(output.size())}};
}}

// ----------------------------------------------------------------------------
// TER codes
// ----------------------------------------------------------------------------

struct TERExportInternal {{
    int code;
    std::string codeStr;
    std::string description;
}};

TERExport
mutateTERcodes(TERExportInternal const& TERcode)
{{
    return TERExport{{
        TERcode.code,
        TERcode.codeStr.c_str(),
        TERcode.description.c_str(),
    }};
}}

extern "C"
Container<TERExport>
getTERcodes()
{{
    static std::vector<TERExportInternal> const codes = []{{
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        std::vector<TERExportInternal> temp = {{}};
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module pluginImport = py::module_::import("{module_name}");
        if (!hasattr(pluginImport, "ter_codes")) {{
            return temp;
        }}
        auto codes = pluginImport.attr("ter_codes").cast<std::vector<py::object>>();
        for (py::object variable: codes)
        {{
            temp.emplace_back(TERExportInternal{{
                variable.attr("code").cast<int>(),
                variable.attr("code_str").cast<std::string>(),
                variable.attr("description").cast<std::string>()}});
        }}
        return temp;
    }}();
    static std::vector<TERExport> output;
    output.reserve(codes.size());
    std::transform(codes.begin(), codes.end(), std::back_inserter(output), mutateTERcodes);
    return {{const_cast<TERExport *>(output.data()), static_cast<int>(output.size())}};
}}

// ----------------------------------------------------------------------------
// Invariant Checks
// ----------------------------------------------------------------------------

static std::map<void*, std::vector<std::string>> checkData;

void
visitEntryExport(
    void* id,
    bool isDelete,
    std::shared_ptr<SLE const> const& before,
    std::shared_ptr<SLE const> const& after)
{{
    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::module_::import("sys").attr("path").attr("append")("{python_folder}");
    py::module pluginImport = py::module_::import("{module_name}");
    if (!hasattr(pluginImport, "invariant_checks")) {{
        throw std::runtime_error("Expected `invariant_checks` variable");
    }}
    auto checks = pluginImport.attr("invariant_checks").cast<std::vector<py::object>>();
    std::vector<std::string> storedCheckData;
    if (auto it = checkData.find(id); it != checkData.end())
    {{
        storedCheckData = it->second;
    }} else
    {{
        for (int i = 0; i < checks.size(); i++)
        {{
            storedCheckData.push_back("");
        }}
        checkData.insert({{id, storedCheckData}});
    }}

    for (int i = 0; i < checks.size(); i++)
    {{
        std::string data = storedCheckData[i];
        py::function checkType = checks[i];
        py::object check = checkType();

        py::object function = check.attr("visit_entry_actual");
        try {{
            py::object ret = function(
                data.length() == 0 ? py::none() : py::cast(data),
                isDelete,
                py::cast(before, py::return_value_policy::reference),
                py::cast(after, py::return_value_policy::reference)
            );
            storedCheckData[i] = ret.cast<std::string>();
        }} catch (py::error_already_set& e) {{
            // Print the error message
            const char* errorMessage = e.what();
            std::cout << "Python Error: " << errorMessage << std::endl;
            throw;
        }}
    }}
    checkData[id] = storedCheckData;
}}

bool
finalizeExport(
    void* id,
    STTx const& tx,
    TER const result,
    XRPAmount const fee,
    ReadView const& view,
    beast::Journal const& j)
{{
    bool finalizeResult = true;

    CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
    py::module_::import("sys").attr("path").attr("append")("{python_folder}");
    py::module pluginImport = py::module_::import("{module_name}");
    if (!hasattr(pluginImport, "invariant_checks")) {{
        throw std::runtime_error("Expected `invariant_checks` variable");
    }}
    auto checks = pluginImport.attr("invariant_checks").cast<std::vector<py::object>>();
    std::vector<std::string> storedCheckData;
    if (auto it = checkData.find(id); it != checkData.end())
    {{
        storedCheckData = it->second;
    }} else
    {{
        JLOG(j.fatal()) << "Invariant failed: could not find matching ID";
        return false;
    }}
    for (int i = 0; i < checks.size(); i++)
    {{
        std::string data = storedCheckData[i];
        py::function checkType = checks[i];
        py::object check = checkType();

        py::object function = check.attr("finalize_actual");
        try {{
            py::object ret = function(
                data,
                py::cast(tx, py::return_value_policy::reference),
                py::cast(result, py::return_value_policy::reference),
                py::cast(fee, py::return_value_policy::reference),
                py::cast(view, py::return_value_policy::reference),
                py::cast(j, py::return_value_policy::reference)
            );
            finalizeResult &= ret.cast<bool>();
        }} catch (py::error_already_set& e) {{
            // Print the error message
            const char* errorMessage = e.what();
            std::cout << "Python Error: " << errorMessage << std::endl;
            throw;
        }}
    }}
    checkData.erase(id);
    return finalizeResult;
}}

extern "C"
Container<InvariantCheckExport>
getInvariantChecks()
{{
    static std::vector<InvariantCheckExport> const checks = []{{
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module_ module = py::module_::import("{module_name}");
        std::vector<InvariantCheckExport> temp = {{}};
        if (!hasattr(module, "invariant_checks")) {{
            return temp;
        }}
        temp.emplace_back(InvariantCheckExport{{
            visitEntryExport,
            finalizeExport}});
        return temp;
    }}();
    return {{const_cast<InvariantCheckExport *>(checks.data()), static_cast<int>(checks.size())}};
}}

// ----------------------------------------------------------------------------
// Inner Object Formats
// ----------------------------------------------------------------------------

struct InnerObjectExportInternal {{
    int code;
    std::string name;
    std::vector<SOElementExport> format;
}};

InnerObjectExport
mutateInnerObject(InnerObjectExportInternal const& innerObject)
{{
    return InnerObjectExport{{
        innerObject.code,
        innerObject.name.c_str(),
        {{const_cast<SOElementExport *>(innerObject.format.data()), static_cast<int>(innerObject.format.size())}},
    }};
}}

extern "C"
Container<InnerObjectExport>
getInnerObjectFormats()
{{
    static std::vector<InnerObjectExportInternal> const innerObjects = []{{
        std::vector<InnerObjectExportInternal> temp = {{}};
        CustomScopedInterpreter guard{{}}; // start the interpreter and keep it alive
        py::module_::import("sys").attr("path").attr("append")("{python_folder}");
        py::module pluginImport = py::module_::import("{module_name}");
        if (!hasattr(pluginImport, "inner_objects")) {{
            return temp;
        }}
        auto innerObjects = pluginImport.attr("inner_objects").cast<std::vector<py::object>>();
        for (int i = 0; i < innerObjects.size(); i++)
        {{
            py::object innerObject = innerObjects[i];

            std::vector<SOElementExport> format{{}};
            auto pythonFormat = innerObject.attr("format").cast<std::vector<py::object>>();
            for (py::object innerObject: pythonFormat)
            {{
                py::tuple tup = innerObject.cast<py::tuple>();
                WSF sfield = tup[0].cast<WSF>();
                SOEStyle varType = tup[1].cast<SOEStyle>();
                format.emplace_back(SOElementExport{{
                    static_cast<SField const&>(sfield).getCode(),
                    varType}});
            }}

            WSF wrappedField = innerObject.attr("field").cast<WSF>();
            SField const& field = static_cast<SField const&>(wrappedField);

            auto exportedTx = InnerObjectExportInternal{{
                field.getCode(),
                std::string(field.jsonName.c_str()),
                format,
            }};

            temp.emplace_back(exportedTx);
        }}
        return temp;
    }}();
    static std::vector<InnerObjectExport> output;
    output.reserve(innerObjects.size());
    std::transform(innerObjects.begin(), innerObjects.end(), std::back_inserter(output), mutateInnerObject);
    return {{const_cast<InnerObjectExport *>(output.data()), static_cast<int>(output.size())}};
}}
"""


def create_files(python_file):
    abs_python_file = os.path.abspath(python_file)
    python_folder = os.path.dirname(abs_python_file)
    sys.path.append(python_folder)
    last_slash = abs_python_file.rfind("/")
    module_name = abs_python_file[(last_slash + 1) : -3]
    # module = importlib.import_module(module_name)
    tx_name = "NewEscrowCreate"  # module.tx_name
    # TODO: add logic to check validity of Python transactors
    # TODO: switch to Jinja
    # TODO: add logic to only generate the methods that exist in Python

    with open(f"{tx_name}.cpp", "w") as f:
        f.write(generate_cpp(tx_name, module_name, python_folder))

    return os.path.abspath(f"{tx_name}.cpp"), module_name


def build_files(cpp_file, project_name):
    with tempfile.TemporaryDirectory() as build_temp:
        build_source_dir = os.path.dirname(__file__)
        conan_source_dir = build_source_dir
        conan_build_dir = os.path.join(conan_source_dir, "build", "generators")
        cmake_args = []
        build_args = []
        if not os.path.exists(conan_build_dir):
            subprocess.run(
                [
                    "conan",
                    "install",
                    conan_source_dir,
                    "--build",
                    "missing",
                    "--settings",
                    "build_type=Debug",
                ],
                check=True,
                cwd=build_temp,
                stdout=subprocess.DEVNULL,
            )
        conan_cmake_file = os.path.join(conan_build_dir, "conan_toolchain.cmake")
        cmake_args += [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={os.getcwd()}{os.sep}",
            f"-DPROJECT_NAME={project_name}",
            f"-DSOURCE_FILE={cpp_file}",
            f"-DCMAKE_TOOLCHAIN_FILE:FILEPATH={conan_cmake_file}",
            "-DCMAKE_BUILD_TYPE=Debug",
            "-DBUILD_SHARED_LIBS=ON",
        ]
        subprocess.run(
            ["cmake", build_source_dir] + cmake_args,
            cwd=build_temp,
            check=True,
            stdout=subprocess.DEVNULL,
        )
        subprocess.run(
            ["cmake", "--build", "."] + build_args,
            cwd=build_temp,
            check=True,
            stdout=subprocess.DEVNULL,
        )


def build():
    cpp_file_fullpath, module_name = create_files(sys.argv[1])
    build_files(cpp_file_fullpath, module_name)



import xrpl_plugin.rippled_py.sfields as rippled_sfields
from xrpl_plugin.rippled_py.sfields import *
from xrpl_plugin.stypes import STBase, STObject, STPluginType

__doc__ = rippled_sfields.__doc__


def create_new_sfield(cls, field_name, field_value):
    if not issubclass(cls, STBase) and cls != STObject:
        # TODO: fix inheritance
        raise Exception("SField must be of an `ST` type.")
    if cls == STPluginType:
        raise Exception("Must use `construct_custom_sfield` for custom STypes.")
    fn_name = f"_create_new_sfield_{cls.__name__}"
    create_fn = getattr(rippled_sfields, fn_name, None)
    if create_fn is None:
        # NOTE: This should never be hit in prod
        # It may be hit during dev work since not everything is implemented yet
        raise Exception(
            f"`_create_new_sfield` function does not exist for {cls.__name__}"
        )
    return create_fn(field_value, field_name)


sf_tick_size = rippled_sfields.sf_tick_size
"""
The number of significant digits in the exchange rate of an offer.
"""


__all__ = list(rippled_sfields.__dict__.keys()) + ["create_new_sfield"]



#!/usr/bin/env python

#    This file is part of rippled: https://github.com/ripple/rippled
#    Copyright (c) 2012 - 2017 Ripple Labs Inc.
#
#    Permission to use, copy, modify, and/or distribute this software for any
#    purpose  with  or without fee is hereby granted, provided that the above
#    copyright notice and this permission notice appear in all copies.
#
#    THE  SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#    WITH  REGARD  TO  THIS  SOFTWARE  INCLUDING  ALL  IMPLIED  WARRANTIES  OF
#    MERCHANTABILITY  AND  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#    ANY  SPECIAL ,  DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#    WHATSOEVER  RESULTING  FROM  LOSS  OF USE, DATA OR PROFITS, WHETHER IN AN
#    ACTION  OF  CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#    OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""
Invocation:

    ./Builds/Test.py - builds and tests all configurations

The build must succeed without shell aliases for this to work.

To pass flags to cmake, put them at the very end of the command line, after
the -- flag - like this:

    ./Builds/Test.py -- -j4  # Pass -j4 to cmake --build


Common problems:

1) Boost not found. Solution: export BOOST_ROOT=[path to boost folder]

2) OpenSSL not found. Solution: export OPENSSL_ROOT=[path to OpenSSL folder]

3) cmake is not found. Solution: Be sure cmake directory is on your $PATH

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import itertools
import os
import platform
import re
import shutil
import sys
import subprocess


def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))

IS_WINDOWS = platform.system().lower() == 'windows'
IS_OS_X = platform.system().lower() == 'darwin'

# CMake
if IS_WINDOWS:
    CMAKE_UNITY_CONFIGS = ['Debug', 'Release']
    CMAKE_NONUNITY_CONFIGS = ['Debug', 'Release']
else:
    CMAKE_UNITY_CONFIGS = []
    CMAKE_NONUNITY_CONFIGS = []
CMAKE_UNITY_COMBOS = { '' : [['rippled'], CMAKE_UNITY_CONFIGS],
    '.nounity' : [['rippled'], CMAKE_NONUNITY_CONFIGS] }

if IS_WINDOWS:
    CMAKE_DIR_TARGETS = { ('msvc' + unity,) : targets for unity, targets in
        CMAKE_UNITY_COMBOS.items() }
elif IS_OS_X:
    CMAKE_DIR_TARGETS = { (build + unity,) : targets
                   for build in ['debug', 'release']
                   for unity, targets in CMAKE_UNITY_COMBOS.items() }
else:
    CMAKE_DIR_TARGETS = { (cc + "." + build + unity,) : targets
                   for cc in ['gcc', 'clang']
                   for build in ['debug', 'release', 'coverage', 'profile']
                   for unity, targets in CMAKE_UNITY_COMBOS.items() }

# list of tuples of all possible options
if IS_WINDOWS or IS_OS_X:
    CMAKE_ALL_GENERATE_OPTIONS = [tuple(x) for x in powerset(['-GNinja', '-Dassert=true'])]
else:
    CMAKE_ALL_GENERATE_OPTIONS = list(set(
        [tuple(x) for x in powerset(['-GNinja', '-Dstatic=true', '-Dassert=true', '-Dsan=address'])] +
        [tuple(x) for x in powerset(['-GNinja', '-Dstatic=true', '-Dassert=true', '-Dsan=thread'])]))

parser = argparse.ArgumentParser(
    description='Test.py - run ripple tests'
)

parser.add_argument(
    '--all', '-a',
    action='store_true',
    help='Build all configurations.',
)

parser.add_argument(
    '--keep_going', '-k',
    action='store_true',
    help='Keep going after one configuration has failed.',
)

parser.add_argument(
    '--silent', '-s',
    action='store_true',
    help='Silence all messages except errors',
)

parser.add_argument(
    '--verbose', '-v',
    action='store_true',
    help=('Report more information about which commands are executed and the '
          'results.'),
)

parser.add_argument(
    '--test', '-t',
    default='',
    help='Add a prefix for unit tests',
)

parser.add_argument(
    '--testjobs',
    default='0',
    type=int,
    help='Run tests in parallel'
)

parser.add_argument(
    '--ipv6',
    action='store_true',
    help='Use IPv6 localhost when running unit tests.',
)

parser.add_argument(
    '--clean', '-c',
    action='store_true',
    help='delete all build artifacts after testing',
)

parser.add_argument(
    '--quiet', '-q',
    action='store_true',
    help='Reduce output where possible (unit tests)',
)

parser.add_argument(
    '--dir', '-d',
    default=(),
    nargs='*',
    help='Specify one or more CMake dir names. '
        'Will also be used as -Dtarget=<dir> running cmake.'
)

parser.add_argument(
    '--target',
    default=(),
    nargs='*',
    help='Specify one or more CMake build targets. '
        'Will be used as --target <target> running cmake --build.'
    )

parser.add_argument(
    '--config',
    default=(),
    nargs='*',
    help='Specify one or more CMake build configs. '
        'Will be used as --config <config> running cmake --build.'
    )

parser.add_argument(
    '--generator_option',
    action='append',
    help='Specify a CMake generator option. Repeat for multiple options. '
        'Will be passed to the cmake generator. '
        'Due to limits of the argument parser, arguments starting with \'-\' '
        'must be attached to this option. e.g. --generator_option=-GNinja.')

parser.add_argument(
    '--build_option',
    action='append',
    help='Specify a build option. Repeat for multiple options. '
        'Will be passed to the build tool via cmake --build. '
        'Due to limits of the argument parser, arguments starting with \'-\' '
        'must be attached to this option. e.g. --build_option=-j8.')

parser.add_argument(
    'extra_args',
    default=(),
    nargs='*',
    help='Extra arguments are passed through to the tools'
)

ARGS = parser.parse_args()

def decodeString(line):
    # Python 2 vs. Python 3
    if isinstance(line, str):
        return line
    else:
        return line.decode()

def shell(cmd, args=(), silent=False, cust_env=None):
    """"Execute a shell command and return the output."""
    silent = ARGS.silent or silent
    verbose = not silent and ARGS.verbose
    if verbose:
        print('$' + cmd, *args)

    command = (cmd,) + args

    # shell is needed in Windows to find executable in the path
    process = subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=cust_env,
        shell=IS_WINDOWS)
    lines = []
    count = 0
    # readline returns '' at EOF
    for line in iter(process.stdout.readline, ''):
        if process.poll() is None:
            decoded = decodeString(line)
            lines.append(decoded)
            if verbose:
                print(decoded, end='')
            elif not silent:
                count += 1
                if count >= 80:
                    print()
                    count = 0
                else:
                    print('.', end='')
        else:
            break

    if not verbose and count:
        print()
    process.wait()
    return process.returncode, lines

def get_cmake_dir(cmake_dir):
    return os.path.join('build' , 'cmake' , cmake_dir)

def run_cmake(directory, cmake_dir, args):
    print('Generating build in', directory, 'with', *args or ('default options',))
    old_dir = os.getcwd()
    if not os.path.exists(directory):
        os.makedirs(directory)
    os.chdir(directory)
    if IS_WINDOWS and not any(arg.startswith("-G") for arg in args) and not os.path.exists("CMakeCache.txt"):
        if '--ninja' in args:
            args += ( '-GNinja', )
        else:
            args += ( '-GVisual Studio 14 2015 Win64', )
    # hack to extract cmake options/args from the legacy target format
    if re.search('\.unity', cmake_dir):
        args += ( '-Dunity=ON', )
    if re.search('\.nounity', cmake_dir):
        args += ( '-Dunity=OFF', )
    if re.search('coverage', cmake_dir):
        args += ( '-Dcoverage=ON', )
    if re.search('profile', cmake_dir):
        args += ( '-Dprofile=ON', )
    if re.search('debug', cmake_dir):
        args += ( '-DCMAKE_BUILD_TYPE=Debug', )
    if re.search('release', cmake_dir):
        args += ( '-DCMAKE_BUILD_TYPE=Release', )
    m = re.search('gcc(-[^.]*)', cmake_dir)
    if m:
        args += ( '-DCMAKE_C_COMPILER=' + m.group(0),
          '-DCMAKE_CXX_COMPILER=g++' + m.group(1), )
    elif re.search('gcc', cmake_dir):
        args += ( '-DCMAKE_C_COMPILER=gcc', '-DCMAKE_CXX_COMPILER=g++', )
    m = re.search('clang(-[^.]*)', cmake_dir)
    if m:
        args += ( '-DCMAKE_C_COMPILER=' + m.group(0),
          '-DCMAKE_CXX_COMPILER=clang++' + m.group(1), )
    elif re.search('clang', cmake_dir):
        args += ( '-DCMAKE_C_COMPILER=clang', '-DCMAKE_CXX_COMPILER=clang++', )

    args += ( os.path.join('..', '..', '..'), )
    resultcode, lines = shell('cmake', args)

    if resultcode:
        print('Generating FAILED:')
        if not ARGS.verbose:
            print(*lines, sep='')
        sys.exit(1)

    os.chdir(old_dir)

def run_cmake_build(directory, target, config, args):
    print('Building', target, config, 'in', directory, 'with', *args or ('default options',))
    build_args=('--build', directory)
    if target:
      build_args += ('--target', target)
    if config:
      build_args += ('--config', config)
    if args:
        build_args += ('--',)
        build_args += tuple(args)
    resultcode, lines = shell('cmake', build_args)

    if resultcode:
        print('Build FAILED:')
        if not ARGS.verbose:
            print(*lines, sep='')
        sys.exit(1)

def run_cmake_tests(directory, target, config):
    failed = []
    if IS_WINDOWS:
        target += '.exe'
    executable = os.path.join(directory, config if config else 'Debug', target)
    if(not os.path.exists(executable)):
        executable = os.path.join(directory, target)
    print('Unit tests for', executable)
    testflag = '--unittest'
    quiet = ''
    testjobs = ''
    ipv6 = ''
    if ARGS.test:
        testflag += ('=' + ARGS.test)
    if ARGS.quiet:
        quiet = '-q'
    if ARGS.ipv6:
        ipv6 = '--unittest-ipv6'
    if ARGS.testjobs:
        testjobs = ('--unittest-jobs=' + str(ARGS.testjobs))
    resultcode, lines = shell(executable, (testflag, quiet, testjobs, ipv6))

    if resultcode:
        if not ARGS.verbose:
            print('ERROR:', *lines, sep='')
        failed.append([target, 'unittest'])

    return failed

def main():
    all_failed = []
    if ARGS.all:
        build_dir_targets = CMAKE_DIR_TARGETS
        generator_options = CMAKE_ALL_GENERATE_OPTIONS
    else:
        build_dir_targets = { tuple(ARGS.dir) : [ARGS.target, ARGS.config] }
        if ARGS.generator_option:
            generator_options = [tuple(ARGS.generator_option)]
        else:
            generator_options = [tuple()]

    if not build_dir_targets:
        # Let CMake choose the build tool.
        build_dir_targets = { () : [] }

    if ARGS.build_option:
        ARGS.build_option = ARGS.build_option + list(ARGS.extra_args)
    else:
        ARGS.build_option = list(ARGS.extra_args)

    for args in generator_options:
        for build_dirs, (build_targets, build_configs) in build_dir_targets.items():
            if not build_dirs:
                build_dirs = ('default',)
            if not build_targets:
                build_targets = ('rippled',)
            if not build_configs:
                build_configs = ('',)
            for cmake_dir in build_dirs:
                cmake_full_dir = get_cmake_dir(cmake_dir)
                run_cmake(cmake_full_dir, cmake_dir, args)

                for target in build_targets:
                    for config in build_configs:
                        run_cmake_build(cmake_full_dir, target, config, ARGS.build_option)
                        failed = run_cmake_tests(cmake_full_dir, target, config)

                        if failed:
                            print('FAILED:', *(':'.join(f) for f in failed))
                            if not ARGS.keep_going:
                                sys.exit(1)
                            else:
                                all_failed.extend([decodeString(cmake_dir +
                                        "." + target + "." + config), ':'.join(f)]
                                    for f in failed)
                        else:
                            print('Success')
                if ARGS.clean:
                    shutil.rmtree(cmake_full_dir)

    if all_failed:
        if len(all_failed) > 1:
            print()
            print('FAILED:', *(':'.join(f) for f in all_failed))
        sys.exit(1)

if __name__ == '__main__':
    main()
    sys.exit(0)



import xrpl_plugin.rippled_py.stypes as rippled_stypes
from xrpl_plugin.rippled_py.stypes import *
from xrpl_plugin.stypes import parser_errors

__doc__ = rippled_stypes.__doc__

__all__ = list(rippled_stypes.__dict__.keys()) + ["parser_errors"]



import xrpl_plugin.rippled_py.ledger_objects as rippled_ledger_objects
from xrpl_plugin.rippled_py.ledger_objects import *

__doc__ = rippled_ledger_objects.__doc__

__all__ = list(rippled_ledger_objects.__dict__.keys())



import xrpl_plugin.rippled_py.stypes.parser_errors as rippled_parser_errors
from xrpl_plugin.rippled_py.stypes.parser_errors import *

__doc__ = rippled_parser_errors.__doc__

__all__ = list(rippled_parser_errors.__dict__.keys())



from setuptools import setup
import os
import subprocess
from pathlib import Path

# Available at setup time due to pyproject.toml
from pybind11.setup_helpers import Pybind11Extension  # , build_ext
from setuptools import find_packages
from setuptools.command.build_ext import build_ext

# The main interface is through Pybind11Extension.
# * You can add cxx_std=11/14/17, and then build_ext can be removed.
# * You can set include_pybind11=false to add the include directory yourself,
#   say from a submodule.
#
# Note:
#   Sort input source files if you glob sources to ensure bit-for-bit
#   reproducible builds (https://github.com/pybind/python_example/pull/53)

ext_modules = [
    Pybind11Extension(
        "xrpl_plugin.rippled_py",
        ["src/main.cpp"],
        cxx_std=17,
    ),
]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class CMakeBuild(build_ext):
    def build_extension(self, ext) -> None:
        # Must be in this form due to bug in .resolve() only fixed in Python 3.10+
        ext_fullpath = Path.cwd() / self.get_ext_fullpath(ext.name)  # type: ignore[no-untyped-call]
        extdir = ext_fullpath.parent.resolve()

        # Using this requires trailing slash for auto-detection & inclusion of
        # auxiliary "native" libs

        # debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        debug = True
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        # cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}{os.sep}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
            "-DBUILD_SHARED_LIBS=YES",
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        # In this example, we pass in the version to C++. You might not need to.
        cmake_args += [f"-DEXAMPLE_VERSION_INFO={self.distribution.get_version()}"]  # type: ignore[attr-defined]

        # if self.compiler.compiler_type != "msvc":
        #     # Using Ninja-build since it a) is available as a wheel and b)
        #     # multithreads automatically. MSVC would require all variables be
        #     # exported for Ninja to pick it up, which is a little tricky to do.
        #     # Users can override the generator with CMAKE_GENERATOR in CMake
        #     # 3.15+.
        #     if not cmake_generator or cmake_generator == "Ninja":
        #         try:
        #             import ninja  # noqa: F401

        #             ninja_executable_path = Path(ninja.BIN_DIR) / "ninja"
        #             cmake_args += [
        #                 "-GNinja",
        #                 f"-DCMAKE_MAKE_PROGRAM:FILEPATH={ninja_executable_path}",
        #             ]
        #         except ImportError:
        #             pass

        # else:

        #     # Single config generators are handled "normally"
        #     single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

        #     # CMake allows an arch-in-generator style for backward compatibility
        #     contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

        #     # Specify the arch if using MSVC generator, but only if it doesn't
        #     # contain a backward-compatibility arch spec already in the
        #     # generator name.
        #     if not single_config and not contains_arch:
        #         cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

        #     # Multi-config generators have a different way to specify configs
        #     if not single_config:
        #         cmake_args += [
        #             f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}"
        #         ]
        #         build_args += ["--config", cfg]

        # if sys.platform.startswith("darwin"):
        #     # Cross-compile support for macOS - respect ARCHFLAGS if set
        #     archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
        #     if archs:
        #         cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += [f"-j{self.parallel}"]

        build_temp = Path(self.build_temp) / ext.name
        if not build_temp.exists():
            build_temp.mkdir(parents=True)

        conan_cmake = "build/generators/conan_toolchain.cmake"
        if not os.path.exists(os.path.join(os.getcwd(), conan_cmake)):
            subprocess.run(
                [
                    "conan",
                    "install",
                    os.getcwd(),
                    "--build",
                    "missing",
                    "--settings",
                    f"build_type={cfg}",
                ],
                cwd=build_temp,
                check=True,
            )
        cmake_args += [f"-DCMAKE_TOOLCHAIN_FILE:FILEPATH={conan_cmake}"]
        subprocess.run(["cmake", os.getcwd()] + cmake_args, cwd=build_temp, check=True)
        subprocess.run(
            ["cmake", "--build", "."] + build_args, cwd=build_temp, check=True
        )


setup(
    name="xrpl-plugin",
    version="0.1.0",
    author="Mayukha Vadari",
    author_email="mvadari@ripple.com",
    url="https://github.com/mvadari/xrpl-plugin",
    description="XRPL Plugins",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    ext_modules=ext_modules,
    # Currently, build_ext only provides an optional "highest supported C++
    # level" feature, but in the future it may provide more features.
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    python_requires=">=3.8",
    entry_points="""
        [console_scripts]
        plugin-build=xrpl_plugin:build
    """,
)



from conan import ConanFile
from conan.tools.build import check_min_cppstd
from conan.tools.cmake import CMake, CMakeDeps, CMakeToolchain, cmake_layout
from conan.tools.files import apply_conandata_patches, copy, export_conandata_patches, get, rmdir
from conan.tools.microsoft import is_msvc
from conan.tools.scm import Version
from conan.errors import ConanInvalidConfiguration
import os

required_conan_version = ">=1.55.0"


class SociConan(ConanFile):
    name = "soci"
    homepage = "https://github.com/SOCI/soci"
    url = "https://github.com/conan-io/conan-center-index"
    description = "The C++ Database Access Library "
    topics = ("mysql", "odbc", "postgresql", "sqlite3")
    license = "BSL-1.0"

    settings = "os", "arch", "compiler", "build_type"
    options = {
        "shared":           [True, False],
        "fPIC":             [True, False],
        "empty":            [True, False],
        "with_sqlite3":     [True, False],
        "with_db2":         [True, False],
        "with_odbc":        [True, False],
        "with_oracle":      [True, False],
        "with_firebird":    [True, False],
        "with_mysql":       [True, False],
        "with_postgresql":  [True, False],
        "with_boost":       [True, False],
    }
    default_options = {
        "shared":           False,
        "fPIC":             True,
        "empty":            False,
        "with_sqlite3":     False,
        "with_db2":         False,
        "with_odbc":        False,
        "with_oracle":      False,
        "with_firebird":    False,
        "with_mysql":       False,
        "with_postgresql":  False,
        "with_boost":       False,
    }

    def export_sources(self):
        export_conandata_patches(self)

    def layout(self):
        cmake_layout(self, src_folder="src")

    def config_options(self):
        if self.settings.os == "Windows":
            self.options.rm_safe("fPIC")

    def configure(self):
        if self.options.shared:
            self.options.rm_safe("fPIC")

    def requirements(self):
        if self.options.with_sqlite3:
            self.requires("sqlite3/3.41.1")
        if self.options.with_odbc and self.settings.os != "Windows":
            self.requires("odbc/2.3.11")
        if self.options.with_mysql:
            self.requires("libmysqlclient/8.0.31")
        if self.options.with_postgresql:
            self.requires("libpq/14.7")
        if self.options.with_boost:
            self.requires("boost/1.81.0")

    @property
    def _minimum_compilers_version(self):
        return {
            "Visual Studio": "14",
            "gcc": "4.8",
            "clang": "3.8",
            "apple-clang": "8.0"
        }

    def validate(self):
        if self.settings.compiler.get_safe("cppstd"):
            check_min_cppstd(self, 11)

        compiler = str(self.settings.compiler)
        compiler_version = Version(self.settings.compiler.version.value)
        if compiler not in self._minimum_compilers_version:
            self.output.warning("{} recipe lacks information about the {} compiler support.".format(self.name, self.settings.compiler))
        elif compiler_version < self._minimum_compilers_version[compiler]:
            raise ConanInvalidConfiguration("{} requires a {} version >= {}".format(self.name, compiler, compiler_version))

        prefix  = "Dependencies for"
        message = "not configured in this conan package."
        if self.options.with_db2:
            # self.requires("db2/0.0.0") # TODO add support for db2
            raise ConanInvalidConfiguration("{} DB2 {} ".format(prefix, message))
        if self.options.with_oracle:
            # self.requires("oracle_db/0.0.0") # TODO add support for oracle
            raise ConanInvalidConfiguration("{} ORACLE {} ".format(prefix, message))
        if self.options.with_firebird:
            # self.requires("firebird/0.0.0") # TODO add support for firebird
            raise ConanInvalidConfiguration("{} firebird {} ".format(prefix, message))

    def source(self):
        get(self, **self.conan_data["sources"][self.version], strip_root=True)

    def generate(self):
        tc = CMakeToolchain(self)

        tc.variables["SOCI_SHARED"] = self.options.shared
        tc.variables["SOCI_STATIC"] = not self.options.shared
        tc.variables["SOCI_TESTS"] = False
        tc.variables["SOCI_CXX11"] = True
        tc.variables["SOCI_EMPTY"] = self.options.empty
        tc.variables["WITH_SQLITE3"] = self.options.with_sqlite3
        tc.variables["WITH_DB2"] = self.options.with_db2
        tc.variables["WITH_ODBC"] = self.options.with_odbc
        tc.variables["WITH_ORACLE"] = self.options.with_oracle
        tc.variables["WITH_FIREBIRD"] = self.options.with_firebird
        tc.variables["WITH_MYSQL"] = self.options.with_mysql
        tc.variables["WITH_POSTGRESQL"] = self.options.with_postgresql
        tc.variables["WITH_BOOST"] = self.options.with_boost
        tc.generate()

        deps = CMakeDeps(self)
        deps.generate()

    def build(self):
        apply_conandata_patches(self)
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def package(self):
        copy(self, "LICENSE_1_0.txt", dst=os.path.join(self.package_folder, "licenses"), src=self.source_folder)

        cmake = CMake(self)
        cmake.install()

        rmdir(self, os.path.join(self.package_folder, "lib", "cmake"))

    def package_info(self):
        self.cpp_info.set_property("cmake_file_name", "SOCI")

        target_suffix = "" if self.options.shared else "_static"
        lib_prefix = "lib" if is_msvc(self) and not self.options.shared else ""
        version = Version(self.version)
        lib_suffix = "_{}_{}".format(version.major, version.minor) if self.settings.os == "Windows" else ""

        # soci_core
        self.cpp_info.components["soci_core"].set_property("cmake_target_name", "SOCI::soci_core{}".format(target_suffix))
        self.cpp_info.components["soci_core"].libs = ["{}soci_core{}".format(lib_prefix, lib_suffix)]
        if self.options.with_boost:
            self.cpp_info.components["soci_core"].requires.append("boost::boost")

        # soci_empty
        if self.options.empty:
            self.cpp_info.components["soci_empty"].set_property("cmake_target_name", "SOCI::soci_empty{}".format(target_suffix))
            self.cpp_info.components["soci_empty"].libs = ["{}soci_empty{}".format(lib_prefix, lib_suffix)]
            self.cpp_info.components["soci_empty"].requires = ["soci_core"]

        # soci_sqlite3
        if self.options.with_sqlite3:
            self.cpp_info.components["soci_sqlite3"].set_property("cmake_target_name", "SOCI::soci_sqlite3{}".format(target_suffix))
            self.cpp_info.components["soci_sqlite3"].libs = ["{}soci_sqlite3{}".format(lib_prefix, lib_suffix)]
            self.cpp_info.components["soci_sqlite3"].requires = ["soci_core", "sqlite3::sqlite3"]

        # soci_odbc
        if self.options.with_odbc:
            self.cpp_info.components["soci_odbc"].set_property("cmake_target_name", "SOCI::soci_odbc{}".format(target_suffix))
            self.cpp_info.components["soci_odbc"].libs = ["{}soci_odbc{}".format(lib_prefix, lib_suffix)]
            self.cpp_info.components["soci_odbc"].requires = ["soci_core"]
            if self.settings.os == "Windows":
                self.cpp_info.components["soci_odbc"].system_libs.append("odbc32")
            else:
                self.cpp_info.components["soci_odbc"].requires.append("odbc::odbc")

        # soci_mysql
        if self.options.with_mysql:
            self.cpp_info.components["soci_mysql"].set_property("cmake_target_name", "SOCI::soci_mysql{}".format(target_suffix))
            self.cpp_info.components["soci_mysql"].libs = ["{}soci_mysql{}".format(lib_prefix, lib_suffix)]
            self.cpp_info.components["soci_mysql"].requires = ["soci_core", "libmysqlclient::libmysqlclient"]

        # soci_postgresql
        if self.options.with_postgresql:
            self.cpp_info.components["soci_postgresql"].set_property("cmake_target_name", "SOCI::soci_postgresql{}".format(target_suffix))
            self.cpp_info.components["soci_postgresql"].libs = ["{}soci_postgresql{}".format(lib_prefix, lib_suffix)]
            self.cpp_info.components["soci_postgresql"].requires = ["soci_core", "libpq::libpq"]

        # TODO: to remove in conan v2 once cmake_find_package* generators removed
        self.cpp_info.names["cmake_find_package"] = "SOCI"
        self.cpp_info.names["cmake_find_package_multi"] = "SOCI"
        self.cpp_info.components["soci_core"].names["cmake_find_package"] = "soci_core{}".format(target_suffix)
        self.cpp_info.components["soci_core"].names["cmake_find_package_multi"] = "soci_core{}".format(target_suffix)
        if self.options.empty:
            self.cpp_info.components["soci_empty"].names["cmake_find_package"] = "soci_empty{}".format(target_suffix)
            self.cpp_info.components["soci_empty"].names["cmake_find_package_multi"] = "soci_empty{}".format(target_suffix)
        if self.options.with_sqlite3:
            self.cpp_info.components["soci_sqlite3"].names["cmake_find_package"] = "soci_sqlite3{}".format(target_suffix)
            self.cpp_info.components["soci_sqlite3"].names["cmake_find_package_multi"] = "soci_sqlite3{}".format(target_suffix)
        if self.options.with_odbc:
            self.cpp_info.components["soci_odbc"].names["cmake_find_package"] = "soci_odbc{}".format(target_suffix)
            self.cpp_info.components["soci_odbc"].names["cmake_find_package_multi"] = "soci_odbc{}".format(target_suffix)
        if self.options.with_mysql:
            self.cpp_info.components["soci_mysql"].names["cmake_find_package"] = "soci_mysql{}".format(target_suffix)
            self.cpp_info.components["soci_mysql"].names["cmake_find_package_multi"] = "soci_mysql{}".format(target_suffix)
        if self.options.with_postgresql:
            self.cpp_info.components["soci_postgresql"].names["cmake_find_package"] = "soci_postgresql{}".format(target_suffix)
            self.cpp_info.components["soci_postgresql"].names["cmake_find_package_multi"] = "soci_postgresql{}".format(target_suffix)



#!/usr/bin/env python
"""A script to test rippled in an infinite loop of start-sync-stop.

- Requires Python 3.7+.
- Can be stopped with SIGINT.
- Has no dependencies outside the standard library.
"""

import sys

assert sys.version_info.major == 3 and sys.version_info.minor >= 7

import argparse
import asyncio
import configparser
import contextlib
import json
import logging
import os
from pathlib import Path
import platform
import subprocess
import time
import urllib.error
import urllib.request

# Enable asynchronous subprocesses on Windows. The default changed in 3.8.
# https://docs.python.org/3.7/library/asyncio-platforms.html#subprocess-support-on-windows
if (platform.system() == 'Windows' and sys.version_info.major == 3
        and sys.version_info.minor < 8):
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

DEFAULT_EXE = 'rippled'
DEFAULT_CONFIGURATION_FILE = 'rippled.cfg'
# Number of seconds to wait before forcefully terminating.
PATIENCE = 120
# Number of contiguous seconds in a sync state to be considered synced.
DEFAULT_SYNC_DURATION = 60
# Number of seconds between polls of state.
DEFAULT_POLL_INTERVAL = 5
SYNC_STATES = ('full', 'validating', 'proposing')


def read_config(config_file):
    # strict = False: Allow duplicate keys, e.g. [rpc_startup].
    # allow_no_value = True: Allow keys with no values. Generally, these
    # instances use the "key" as the value, and the section name is the key,
    # e.g. [debug_logfile].
    # delimiters = ('='): Allow ':' as a character in Windows paths. Some of
    # our "keys" are actually values, and we don't want to split them on ':'.
    config = configparser.ConfigParser(
        strict=False,
        allow_no_value=True,
        delimiters=('='),
    )
    config.read(config_file)
    return config


def to_list(value, separator=','):
    """Parse a list from a delimited string value."""
    return [s.strip() for s in value.split(separator) if s]


def find_log_file(config_file):
    """Try to figure out what log file the user has chosen. Raises all kinds
    of exceptions if there is any possibility of ambiguity."""
    config = read_config(config_file)
    values = list(config['debug_logfile'].keys())
    if len(values) < 1:
        raise ValueError(
            f'no [debug_logfile] in configuration file: {config_file}')
    if len(values) > 1:
        raise ValueError(
            f'too many [debug_logfile] in configuration file: {config_file}')
    return values[0]


def find_http_port(config_file):
    config = read_config(config_file)
    names = list(config['server'].keys())
    for name in names:
        server = config[name]
        if 'http' in to_list(server.get('protocol', '')):
            return int(server['port'])
    raise ValueError(f'no server in [server] for "http" protocol')


@contextlib.asynccontextmanager
async def rippled(exe=DEFAULT_EXE, config_file=DEFAULT_CONFIGURATION_FILE):
    """A context manager for a rippled process."""
    # Start the server.
    process = await asyncio.create_subprocess_exec(
        str(exe),
        '--conf',
        str(config_file),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    logging.info(f'rippled started with pid {process.pid}')
    try:
        yield process
    finally:
        # Ask it to stop.
        logging.info(f'asking rippled (pid: {process.pid}) to stop')
        start = time.time()
        process.terminate()

        # Wait nicely.
        try:
            await asyncio.wait_for(process.wait(), PATIENCE)
        except asyncio.TimeoutError:
            # Ask the operating system to kill it.
            logging.warning(f'killing rippled ({process.pid})')
            try:
                process.kill()
            except ProcessLookupError:
                pass

        code = await process.wait()
        end = time.time()
        logging.info(
            f'rippled stopped after {end - start:.1f} seconds with code {code}'
        )


async def sync(
        port,
        *,
        duration=DEFAULT_SYNC_DURATION,
        interval=DEFAULT_POLL_INTERVAL,
):
    """Poll rippled on an interval until it has been synced for a duration."""
    start = time.perf_counter()
    while (time.perf_counter() - start) < duration:
        await asyncio.sleep(interval)

        request = urllib.request.Request(
            f'http://127.0.0.1:{port}',
            data=json.dumps({
                'method': 'server_state'
            }).encode(),
            headers={'Content-Type': 'application/json'},
        )
        with urllib.request.urlopen(request) as response:
            try:
                body = json.loads(response.read())
            except urllib.error.HTTPError as cause:
                logging.warning(f'server_state returned not JSON: {cause}')
                start = time.perf_counter()
                continue

        try:
            state = body['result']['state']['server_state']
        except KeyError as cause:
            logging.warning(f'server_state response missing key: {cause.key}')
            start = time.perf_counter()
            continue
        logging.info(f'server_state: {state}')
        if state not in SYNC_STATES:
            # Require a contiguous sync state.
            start = time.perf_counter()


async def loop(test,
               *,
               exe=DEFAULT_EXE,
               config_file=DEFAULT_CONFIGURATION_FILE):
    """
    Start-test-stop rippled in an infinite loop.

    Moves log to a different file after each iteration.
    """
    log_file = find_log_file(config_file)
    id = 0
    while True:
        logging.info(f'iteration: {id}')
        async with rippled(exe, config_file) as process:
            start = time.perf_counter()
            exited = asyncio.create_task(process.wait())
            tested = asyncio.create_task(test())
            # Try to sync as long as the process is running.
            done, pending = await asyncio.wait(
                {exited, tested},
                return_when=asyncio.FIRST_COMPLETED,
            )
            if done == {exited}:
                code = exited.result()
                logging.warning(
                    f'server halted for unknown reason with code {code}')
            else:
                assert done == {tested}
                assert tested.exception() is None
            end = time.perf_counter()
            logging.info(f'synced after {end - start:.0f} seconds')
        os.replace(log_file, f'debug.{id}.log')
        id += 1


logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
)

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument(
    'rippled',
    type=Path,
    nargs='?',
    default=DEFAULT_EXE,
    help='Path to rippled.',
)
parser.add_argument(
    '--conf',
    type=Path,
    default=DEFAULT_CONFIGURATION_FILE,
    help='Path to configuration file.',
)
parser.add_argument(
    '--duration',
    type=int,
    default=DEFAULT_SYNC_DURATION,
    help='Number of contiguous seconds required in a synchronized state.',
)
parser.add_argument(
    '--interval',
    type=int,
    default=DEFAULT_POLL_INTERVAL,
    help='Number of seconds to wait between polls of state.',
)
args = parser.parse_args()

port = find_http_port(args.conf)


def test():
    return sync(port, duration=args.duration, interval=args.interval)


try:
    asyncio.run(loop(test, exe=args.rippled, config_file=args.conf))
except KeyboardInterrupt:
    # Squelch the message. This is a normal mode of exit.
    pass
